@@ -72,6 +72,62 @@ count_from_word(bitarrayobject *a, Py_ssize_t i)
     return popcnt_words(WBUFF(a) + i, a->nbits / 64 - i) + popcnt_64(zlw(a));
 }
 
+/* basically resize() but without importing and exporting buffer checks */
+static int
+resize_lite(bitarrayobject *self, Py_ssize_t nbits)
+{
+    const size_t size = Py_SIZE(self);
+    const size_t allocated = self->allocated;
+    const size_t newsize = BYTES((size_t) nbits);
+    size_t new_allocated;
+
+    assert(allocated >= size && size == BYTES((size_t) self->nbits));
+    assert(self->readonly == 0);
+    assert(self->ob_exports == 0);
+    assert(self->buffer == NULL);
+
+    if (newsize == size) {
+        self->nbits = nbits;
+        return 0;
+    }
+
+    if (newsize == 0) {
+        PyMem_Free(self->ob_item);
+        self->ob_item = NULL;
+        Py_SET_SIZE(self, 0);
+        self->allocated = 0;
+        self->nbits = 0;
+        return 0;
+    }
+
+    if (allocated >= newsize) {
+        if (newsize >= allocated / 2) {
+            Py_SET_SIZE(self, newsize);
+            self->nbits = nbits;
+            return 0;
+        }
+        new_allocated = newsize;
+    }
+    else {
+        new_allocated = newsize;
+        if (size != 0 && newsize / 2 <= allocated) {
+            new_allocated += (newsize >> 4) + (newsize < 8 ? 3 : 7);
+            new_allocated &= ~(size_t) 3;
+        }
+    }
+
+    assert(new_allocated >= newsize);
+    self->ob_item = PyMem_Realloc(self->ob_item, new_allocated);
+    if (self->ob_item == NULL) {
+        PyErr_NoMemory();
+        return -1;
+    }
+    Py_SET_SIZE(self, newsize);
+    self->allocated = new_allocated;
+    self->nbits = nbits;
+    return 0;
+}
+
 /* ---------------------------- zeros / ones --------------------------- */
 
 static PyObject *
@@ -510,47 +566,71 @@ hex_to_int(char c)
     return -1;
 }
 
+static int
+is_whitespace(char c)
+{
+    if (c)
+        return strchr(" \n\r\t\v", c) ? 1 : 0;
+    else  /* NUL character is not whitespace */
+        return 0;
+}
+
 /* create hexadecimal string from bitarray */
 static char *
-ba2hex_core(bitarrayobject *a)
+ba2hex_core(bitarrayobject *a, Py_ssize_t group, char *sep)
 {
-    const int le = IS_LE(a), be = IS_BE(a);
-    const size_t strsize = a->nbits / 4;
+    const int be = IS_BE(a);
+    size_t strsize = a->nbits / 4, j, nsep;
+    Py_ssize_t i;
     char *buff = a->ob_item, *str;
-    size_t i;
 
-    assert(a->nbits % 4 == 0);
+    nsep = (group && strsize) ? strlen(sep) : 0;  /* 0 means no grouping */
+    if (nsep)
+        strsize += nsep * ((strsize - 1) / group);
 
-    str = (char *) PyMem_Malloc(strsize + 1);
-    if (str == NULL)
+    if ((str = (char *) PyMem_Malloc(strsize + 1)) == NULL)
         return NULL;
 
-    /* translate entire bitarray buffer, even when we have 4 pad bits */
-    for (i = 0; i < strsize; i += 2) {
-        unsigned char c = *buff++;
-        str[i + le] = hexdigits[c >> 4];
-        str[i + be] = hexdigits[0x0f & c];
+    for (i = j = 0; i < a->nbits / 4; i++) {
+        unsigned char c = buff[i / 2];
+
+        if (nsep && i && i % group == 0) {
+            memcpy(str + j, sep, nsep);
+            j += nsep;
+        }
+        str[j++] = hexdigits[(i + be) % 2 ? c >> 4 : 0x0f & c];
     }
+    assert(j == strsize);
     str[strsize] = 0;  /* terminate string */
     return str;
 }
 
 static PyObject *
-ba2hex(PyObject *module, PyObject *obj)
+ba2hex(PyObject *module, PyObject *args, PyObject *kwds)
 {
+    static char *kwlist[] = {"", "group", "sep", NULL};
     PyObject *result;
     bitarrayobject *a;
-    char *str;
+    Py_ssize_t group = 0;
+    char *sep = " ", *str;
 
-    if (ensure_bitarray(obj) < 0)
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!|ns:ba2hex", kwlist,
+                                     bitarray_type_obj, (PyObject *) &a,
+                                     &group, &sep))
         return NULL;
 
-    a = (bitarrayobject *) obj;
-    if (a->nbits % 4)
-        return PyErr_Format(PyExc_ValueError, "bitarray length %zd not "
-                            "multiple of 4", a->nbits);
+    if (a->nbits % 4) {
+        PyErr_Format(PyExc_ValueError, "bitarray length %zd not "
+                     "multiple of 4", a->nbits);
+        return NULL;
+    }
+    if (group < 0) {
+        PyErr_Format(PyExc_ValueError, "non-negative integer "
+                     "expected for group, got: %zd", group);
+        return NULL;
+    }
 
-    if ((str = ba2hex_core(a)) == NULL)
+    if ((str = ba2hex_core(a, group, sep)) == NULL)
         return PyErr_NoMemory();
 
     result = PyUnicode_FromString(str);
@@ -559,40 +639,44 @@ ba2hex(PyObject *module, PyObject *obj)
 }
 
 PyDoc_STRVAR(ba2hex_doc,
-"ba2hex(bitarray, /) -> hexstr\n\
+"ba2hex(bitarray, /, group=0, sep=' ') -> hexstr\n\
 \n\
 Return a string containing the hexadecimal representation of\n\
-the bitarray (which has to be multiple of 4 in length).");
+the bitarray (which has to be multiple of 4 in length).\n\
+When grouped, the string `sep` is inserted between groups\n\
+of `group` characters, default is a space.");
 
 
-/* Translate hexadecimal digits from 'hexstr' into the bitarray 'a' buffer.
+/* Translate hexadecimal digits from 'hexstr' into the bitarray 'a' buffer,
+   which must be initalized to zeros.
    Each digit corresponds to 4 bits in the bitarray.
    Note that the number of hexadecimal digits may be odd. */
 static int
 hex2ba_core(bitarrayobject *a, Py_buffer hexstr)
 {
-    const char *str = hexstr.buf;
     const int be = IS_BE(a);
-    Py_ssize_t i;
+    const char *str = hexstr.buf;
+    Py_ssize_t i = 0, j;
 
     assert(a->nbits == 4 * hexstr.len);
 
-    if (a->ob_item)
-        memset(a->ob_item, 0, Py_SIZE(a));
-
-    for (i = 0; i < hexstr.len; i++) {
-        unsigned char c = str[i];
+    for (j = 0; j < hexstr.len; j++) {
+        unsigned char c = str[j];
         int x = hex_to_int(c);
 
         if (x < 0) {
-            PyErr_Format(PyExc_ValueError, "non-hexadecimal digit found, "
-                         "got '%c' (0x%02x)", c, c);
+            if (is_whitespace(c))
+                continue;
+            PyErr_Format(PyExc_ValueError, "invalid digit found for "
+                         "base16, got '%c' (0x%02x)", c, c);
             return -1;
         }
         assert(0 <= x && x < 16);
         a->ob_item[i / 2] |= x << 4 * ((i + be) % 2);
+        i++;
     }
-    return 0;
+    assert(i <= a->nbits);
+    return resize_lite(a, 4 * i);  /* in case we ignored whitespace */
 }
 
 static PyObject *
@@ -607,7 +691,7 @@ hex2ba(PyObject *module, PyObject *args, PyObject *kwds)
                                      &hexstr, &endian))
         return NULL;
 
-    a = new_bitarray(4 * hexstr.len, endian, -1);
+    a = new_bitarray(4 * hexstr.len, endian, 0);
     if (a == NULL)
         goto error;
 
@@ -627,49 +711,41 @@ PyDoc_STRVAR(hex2ba_doc,
 "hex2ba(hexstr, /, endian=None) -> bitarray\n\
 \n\
 Bitarray of hexadecimal representation.  hexstr may contain any number\n\
-(including odd numbers) of hex digits (upper or lower case).");
+(including odd numbers) of hex digits (upper or lower case).\n\
+Whitespace is ignored.");
 
 /* ----------------------- base 2, 4, 8, 16, 32, 64 -------------------- */
 
 /* RFC 4648 Base32 alphabet */
 static const char base32_alphabet[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567";
 
-/* standard base 64 alphabet - also described on RFC 4648 */
+/* standard base 64 alphabet - also described in RFC 4648 */
 static const char base64_alphabet[] =
     "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
 static int
 digit_to_int(int n, char c)
 {
+    static signed char table[2][256];
+    static int setup = 0;
     int i;
 
-    switch (n) {
-    case 32:                    /* base 32 */
-        if ('A' <= c && c <= 'Z')
-            return c - 'A';
-        if ('2' <= c && c <= '7')
-            return c - '2' + 26;
-        break;
-
-    case 64:                    /* base 64 */
-        if ('A' <= c && c <= 'Z')
-            return c - 'A';
-        if ('a' <= c && c <= 'z')
-            return c - 'a' + 26;
-        if ('0' <= c && c <= '9')
-            return c - '0' + 52;
-        if (c == '+')
-            return 62;
-        if (c == '/')
-            return 63;
-        break;
-
-    default:                    /* base 2, 4, 8, 16 */
+    if (n <= 16) {                               /* base 2, 4, 8, 16 */
         i = hex_to_int(c);
         if (i < n)
             return i;
+        return -1;
     }
-    return -1;
+    if (!setup) {
+        for (i = 0; i < 256; i++)
+            table[0][i] = table[1][i] = -1;
+        for (i = 0; i < 32; i++)
+            table[0][(unsigned char) base32_alphabet[i]] = i;
+        for (i = 0; i < 64; i++)
+            table[1][(unsigned char) base64_alphabet[i]] = i;
+        setup = 1;
+    }
+    return table[n / 64][(unsigned char) c];     /* base 32, 64 */
 }
 
 /* return m = log2(n) for m in [1..6] */
@@ -689,12 +765,12 @@ base_to_length(int n)
 
 /* create ASCII string from bitarray and base length m */
 static char *
-ba2base_core(bitarrayobject *a, int m)
+ba2base_core(bitarrayobject *a, int m, Py_ssize_t group, char *sep)
 {
     const int le = IS_LE(a);
-    const size_t strsize = a->nbits / m;
     const char *alphabet;
-    size_t i = 0, j;
+    size_t strsize = a->nbits / m, j, nsep;
+    Py_ssize_t i;
     char *str;
 
     assert(1 <= m && m <= 6 && a->nbits % m == 0);
@@ -705,43 +781,65 @@ ba2base_core(bitarrayobject *a, int m)
     default: alphabet = hexdigits;
     }
 
-    str = (char *) PyMem_Malloc(strsize + 1);
-    if (str == NULL)
+    nsep = (group && strsize) ? strlen(sep) : 0;  /* 0 means no grouping */
+    if (nsep)
+        strsize += nsep * ((strsize - 1) / group);
+
+    if ((str = (char *) PyMem_Malloc(strsize + 1)) == NULL)
         return NULL;
 
-    for (j = 0; j < strsize; j++) {
+    for (i = j = 0; i < a->nbits / m; i++) {
         int k, x = 0;
 
+        if (nsep && i && i % group == 0) {
+            memcpy(str + j, sep, nsep);
+            j += nsep;
+        }
         for (k = 0; k < m; k++) {
             int q = le ? k : (m - k - 1);
-            x |= getbit(a, i++) << q;
+            x |= getbit(a, i * m + k) << q;
         }
-        str[j] = alphabet[x];
+        str[j++] = alphabet[x];
     }
+    assert(j == strsize);
     str[strsize] = 0;  /* terminate string */
     return str;
 }
 
 static PyObject *
-ba2base(PyObject *module, PyObject *args)
+ba2base(PyObject *module, PyObject *args, PyObject *kwds)
 {
+    static char *kwlist[] = {"", "", "group", "sep", NULL};
     bitarrayobject *a;
     PyObject *result;
-    char *str;
+    Py_ssize_t group = 0;
+    char *sep = " ", *str;
     int n, m;
 
-    if (!PyArg_ParseTuple(args, "iO!:ba2base", &n,
-                          bitarray_type_obj, (PyObject *) &a))
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "iO!|ns:ba2base", kwlist,
+                                     &n, bitarray_type_obj, (PyObject *) &a,
+                                     &group, &sep))
         return NULL;
 
     if ((m = base_to_length(n)) < 0)
         return NULL;
 
-    if (a->nbits % m)
-        return PyErr_Format(PyExc_ValueError,
-                            "bitarray length must be multiple of %d", m);
+    if (a->nbits % m) {
+        PyErr_Format(PyExc_ValueError, "bitarray length %zd not "
+                     "multiple of %d", a->nbits, m);
+        return NULL;
+    }
+    if (group < 0) {
+        PyErr_Format(PyExc_ValueError, "non-negative integer "
+                     "expected for group, got: %zd", group);
+        return NULL;
+    }
+
+    if (m == 4)
+        str = ba2hex_core(a, group, sep);
+    else
+        str = ba2base_core(a, m, group, sep);
 
-    str = m == 4 ? ba2hex_core(a) : ba2base_core(a, m);
     if (str == NULL)
         return PyErr_NoMemory();
 
@@ -751,13 +849,15 @@ ba2base(PyObject *module, PyObject *args)
 }
 
 PyDoc_STRVAR(ba2base_doc,
-"ba2base(n, bitarray, /) -> str\n\
+"ba2base(n, bitarray, /, group=0, sep=' ') -> str\n\
 \n\
 Return a string containing the base `n` ASCII representation of\n\
 the bitarray.  Allowed values for `n` are 2, 4, 8, 16, 32 and 64.\n\
 The bitarray has to be multiple of length 1, 2, 3, 4, 5 or 6 respectively.\n\
 For `n=32` the RFC 4648 Base32 alphabet is used, and for `n=64` the\n\
-standard base 64 alphabet is used.");
+standard base 64 alphabet is used.\n\
+When grouped, the string `sep` is inserted between groups\n\
+of `group` characters, default is a space.");
 
 
 /* translate ASCII digits (with base length m) into bitarray buffer */
@@ -775,16 +875,20 @@ base2ba_core(bitarrayobject *a, Py_buffer asciistr, int m)
         int k, x = digit_to_int(n, c);
 
         if (x < 0) {
+            if (is_whitespace(c))
+                continue;
             PyErr_Format(PyExc_ValueError, "invalid digit found for "
-                         "base %d, got '%c' (0x%02x)", n, c, c);
+                         "base%d, got '%c' (0x%02x)", n, c, c);
             return -1;
         }
+        assert(0 <= x && x < n);
         for (k = 0; k < m; k++) {
             int q = le ? k : (m - k - 1);
             setbit(a, i++, x & (1 << q));
         }
     }
-    return 0;
+    assert(i <= a->nbits);
+    return resize_lite(a, i);  /* in case we ignored whitespace */
 }
 
 static PyObject *
@@ -803,11 +907,11 @@ base2ba(PyObject *module, PyObject *args, PyObject *kwds)
     if ((m = base_to_length(n)) < 0)
         goto error;
 
-    a = new_bitarray(m * asciistr.len, endian, -1);
+    a = new_bitarray(m * asciistr.len, endian, m == 4 ? 0 : -1);
     if (a == NULL)
         goto error;
 
-    t = m == 4 ? hex2ba_core(a, asciistr) : base2ba_core(a, asciistr, m);
+    t = (m == 4) ? hex2ba_core(a, asciistr) : base2ba_core(a, asciistr, m);
     if (t < 0)
         goto error;
 
@@ -826,66 +930,10 @@ PyDoc_STRVAR(base2ba_doc,
 Bitarray of base `n` ASCII representation.\n\
 Allowed values for `n` are 2, 4, 8, 16, 32 and 64.\n\
 For `n=32` the RFC 4648 Base32 alphabet is used, and for `n=64` the\n\
-standard base 64 alphabet is used.");
+standard base 64 alphabet is used.  Whitespace is ignored.");
 
 /* ------------------------ utility C functions ------------------------ */
 
-/* like resize() */
-static int
-resize_lite(bitarrayobject *self, Py_ssize_t nbits)
-{
-    const size_t size = Py_SIZE(self);
-    const size_t allocated = self->allocated;
-    const size_t newsize = BYTES((size_t) nbits);
-    size_t new_allocated;
-
-    assert(allocated >= size && size == BYTES((size_t) self->nbits));
-    assert(self->readonly == 0);
-    assert(self->ob_exports == 0);
-    assert(self->buffer == NULL);
-
-    if (newsize == size) {
-        self->nbits = nbits;
-        return 0;
-    }
-
-    if (newsize == 0) {
-        PyMem_Free(self->ob_item);
-        self->ob_item = NULL;
-        Py_SET_SIZE(self, 0);
-        self->allocated = 0;
-        self->nbits = 0;
-        return 0;
-    }
-
-    if (allocated >= newsize) {
-        if (newsize >= allocated / 2) {
-            Py_SET_SIZE(self, newsize);
-            self->nbits = nbits;
-            return 0;
-        }
-        new_allocated = newsize;
-    }
-    else {
-        new_allocated = newsize;
-        if (size != 0 && newsize / 2 <= allocated) {
-            new_allocated += (newsize >> 4) + (newsize < 8 ? 3 : 7);
-            new_allocated &= ~(size_t) 3;
-        }
-    }
-
-    assert(new_allocated >= newsize);
-    self->ob_item = PyMem_Realloc(self->ob_item, new_allocated);
-    if (self->ob_item == NULL) {
-        PyErr_NoMemory();
-        return -1;
-    }
-    Py_SET_SIZE(self, newsize);
-    self->allocated = new_allocated;
-    self->nbits = nbits;
-    return 0;
-}
-
 /* Consume one byte from iteratior and return it's value as an integer
    in range(256).  On failure, set an exception and return -1.  */
 static int
@@ -1924,10 +1972,12 @@ static PyMethodDef module_functions[] = {
     {"deserialize",
                   (PyCFunction) deserialize,
                                            METH_O,       deserialize_doc},
-    {"ba2hex",    (PyCFunction) ba2hex,    METH_O,       ba2hex_doc},
+    {"ba2hex",    (PyCFunction) ba2hex,    METH_KEYWORDS |
+                                           METH_VARARGS, ba2hex_doc},
     {"hex2ba",    (PyCFunction) hex2ba,    METH_KEYWORDS |
                                            METH_VARARGS, hex2ba_doc},
-    {"ba2base",   (PyCFunction) ba2base,   METH_VARARGS, ba2base_doc},
+    {"ba2base",   (PyCFunction) ba2base,   METH_KEYWORDS |
+                                           METH_VARARGS, ba2base_doc},
     {"base2ba",   (PyCFunction) base2ba,   METH_KEYWORDS |
                                            METH_VARARGS, base2ba_doc},
     {"sc_encode", (PyCFunction) sc_encode, METH_O,       sc_encode_doc},
